/************************************************************************/
/*                                                                      */
/*         XX      XX W           W  III  NN    N    GGGG               */
/*           XX  XX    W         W    I   N N   N   G                   */
/*             XX       W   W   W     I   N  N  N   G  GGG              */
/*           XX  XX      W W W W      I   N   N N   G    G              */
/*         XX      XX     W   W      III  N    NN    GGGG               */
/*                                                                      */
/*  QtAirfoil Version 0.1        U.Lehnert                   3 /2010    */
/*                                                                      */
/*  Berechnung der Skelettlinie von Tragflügel-Profilen,                */
/*  deren aerodynamischen Kenngrößen                                    */
/*  und deren Unterteilung in Segmente zur Benutzung mit XWing          */
/*                                                                      */
/************************************************************************/


/*

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "airfoil.h"
#include "linalgebra.h"
// include <stdio.h>      // only for debugging
#include <math.h>

Camberline::Camberline(int NumberOfPanels,
               double LED,
               double TED)
{
  double xmin, xmax;
  double ytan;
  int i;
  
  npan = NumberOfPanels;
  xs = new double[npan+1];
  ys = new double[npan+1];
  xmin=atan(-LED);
  xmax=atan(TED);
  xs[0] = 0.0;
  xs[npan] = 1.0;
  ys[0] = 0.0;
  ys[npan] = 0.0;
  for (i=1; i<npan; i++) {
    ytan = (double)i/npan*(TED+LED)-LED;
    xs[i] = (atan(ytan)-xmin)/(xmax-xmin);
    ys[i] = 0.0;
  }
  aoa = new double[npan];
  for (i=0; i<npan; i++) aoa[i]=0.0;
}

Camberline::~Camberline()
{
  delete xs;
  delete ys;
  delete aoa;
}

int Camberline::npoints()
{
  return(npan+1);
}

double Camberline::xpoint(int i)
{
  double r;
  r = 0.0;
  if ((i>=0) && (i<=npan)) r=xs[i];
  return(r);
}

double Camberline::ypoint(int i)
{
  double r;
  r = 0.0;
  if ((i>=0) && (i<=npan)) r=ys[i];
  return(r);
}

void Camberline::sety(int i, double y)
{
  if ((i>=0) && (i<=npan))
    ys[i] = y;
}

void Camberline::computePoints(Airfoil *af)
{
  int i, n;
  double *yp = new double[10];
  double ycl;
  ys[0]=0.0;
  ys[npan]=0.0;
  for (i=1; i<npan; i++) {
    ycl=0.0;
    n=af->splx->findInverse(xs[i], yp, 10);
    if (n==2) ycl=0.5*(af->sply->f(yp[0])+af->sply->f(yp[1]));
    ys[i]=ycl;
  }
}

void Camberline::computeCoeff(double *cl0, double *alfa0,
                              double* cm0, double *dclda,
                              double* asf, double *clsf)
{
  int i, k;
  double *xvortex = new double[npan];
  double *xcontrol = new double[npan];
  double *gamma = new double[npan];
  double *w = new double[npan];
  double gamma0;

  for (i=0; i<npan; i++) {
    xvortex[i] = 0.75*xs[i] + 0.25*xs[i+1];
    xcontrol[i] = 0.25*xs[i] + 0.75*xs[i+1];
    aoa[i] = atan(-(ys[i+1]-ys[i])/(xs[i+1]-xs[i]));
    // printf(" dx=%9.6f  a=%9.6f deg\n",xs[i+1]-xs[i],aoa[i]*180.0/M_PI);
  }
  // the influence matrix is computed
  // the induced velocity of any vortex is proportional
  // to the inverse distance to the respective control point
  double *indu = new double[npan*npan];
  for (i=0; i<npan; i++)          // lines (control points)
    for (k=0; k<npan; k++)        // columns (vortices)
      indu[npan*i+k] = 1.0/(xcontrol[i]-xvortex[k]);
  // the influence matrix is inversed
  MatrixInverse(indu,npan);
  // compute vortex strengths at zero angle of attack
  MatrVecProd(indu,aoa,gamma,npan);
  *cl0=0.0;
  *cm0=0.0;
  for (i=0; i<npan; i++) {
    gamma[i]*=2.0*M_PI;
    *cl0+=2.0*gamma[i];
    *cm0+=2.0*gamma[i]*(0.25-xvortex[i]);
    // printf(" gamma=%9.6f\n",gamma[i]);
  };
  gamma0=gamma[0];
  // compute slope of lift coeffizient
  for (i=0; i<npan; i++) w[i]=1.0;
  MatrVecProd(indu,w,gamma,npan);
  *dclda=0.0;
  for (i=0; i<npan; i++) {
    gamma[i]*=2.0*M_PI;
    *dclda+=2.0*gamma[i];
  };
  *alfa0 = -*cl0 / *dclda;
  // suction-free entrance
  // zero vorticity at leading edge
  *asf = -gamma0 / gamma[0];
  *clsf = *cl0 + *dclda * *asf;

  delete xvortex;
  delete xcontrol;
  delete indu;
  delete w;
}
 
Airfoil::Airfoil(int NumberOfPoints,
          double *x,
          double *y)
{
  np = NumberOfPoints;
  xp = new double[np];
  yp = new double[np];
  for (int i=0; i<np; i++) {
    xp[i]=x[i];
    yp[i]=y[i];
  };
  splx = new Spline(np,xp);
  sply = new Spline(np,yp);
  outline = new Spline2D(splx,sply);
}

Airfoil::~Airfoil()
{
  delete xp;
  delete yp;
  delete splx;
  delete sply;
  delete outline;
}

int Airfoil::npoints()
{
  return(np);
}

double Airfoil::px(int i)
{
  if ((i>=0) && (i<np))
    return(xp[i]);
  else
    return(0.0);
}

double Airfoil::py(int i)
{
  if ((i>=0) && (i<np))
    return(yp[i]);
  else
    return(0.0);
}

void Airfoil::normalize()
{
  int i, nex;
  double noset, nosex, nosey;
  double r, xte, yte;

  double *te = new double[2*np];
  nex = splx->findExtrema(te, 2*np);
  // printf("x coordinate spline has %d extrema\n",nex);
  
  nosex=1.0;
  noset=0.0;
  nosey=0.0;
  for (i=0; i<nex; i++) {
    if (splx->f(te[i])<nosex) {
      noset=te[i];
      nosex=splx->f(te[i]);
      nosey=sply->f(te[i]);
    };
  };
  // printf("nose :  t=%9.6f  x=%9.6f  y=%9.6f\n\n",noset,nosex,nosey);
  
  // printf("p(0)    :  x=%9.6f  y=%9.6f\n",xp[0],yp[0]);
  // printf("p(np-1) :  x=%9.6f  y=%9.6f\n",xp[np-1],yp[np-1]);
  xte=xp[0];
  if (xp[np-1]>xte) {
    xte=xp[np-1];
    r=(xte-xp[1])/(xp[0]-xp[1]);
    xp[0]=xp[1]+r*(xp[0]-xp[1]);
    yp[0]=yp[1]+r*(yp[0]-yp[1]);
  } else {
    r=(xte-xp[np-2])/(xp[np-1]-xp[np-2]);
    xp[np-1]=xp[np-2]+r*(xp[np-1]-xp[np-2]);
    yp[np-1]=yp[np-2]+r*(yp[np-1]-yp[np-2]);
  };
  yte=(yp[0]+yp[np-1])*0.5;
  // printf("te      :  x=%9.6f  y=%9.6f\n",xte,yte);
  // printf("p(0)    :  x=%9.6f  y=%9.6f\n",xp[0],yp[0]);
  // printf("p(np-1) :  x=%9.6f  y=%9.6f\n\n",xp[np-1],yp[np-1]);
  
  for (i=0; i<np; i++) {
    xp[i]-=nosex;
    xp[i]/=(xte-nosex);
    yp[i]-=nosey;
    yp[i]-=(yte-nosey)*xp[i];
  };
  
  delete splx;
  delete sply;
  delete outline;
  splx = new Spline(np,xp);
  sply = new Spline(np,yp);
  outline = new Spline2D(splx,sply);

}
